# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import time
from enum import Enum

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__credits__ = ["DESY P11"]
__license__ = "LGPLv3+"
__category__ = "General"


class P11DetectorCover(AbstractShutter):
    """
    P11DetectorCover define interface to Tango Detector Cover at DESY P11

    Similar to P11 Shutter tango - but differs
    """

    default_timeout = 4
    move_time_min = 1

    def __init__(self, name):
        super().__init__(name)

        self.simulation = False
        self.simulated_opened = False
        self.simulated_moving = False

        self.cmd_open = None
        self.cmd_close = None
        self.cmd_started = 0

        self.chan_state_open = None
        self.chan_state_close = None

        self.cover_is_open = False
        self.cover_is_closed = False

    def init(self):
        """Initialise the predefined values"""

        self._initialise_values()
        self.simulation = self.get_property("simulation")
        self.cmd_timeout = self.get_property("command_timeout", self.default_timeout)

        if not self.simulation:
            self.cmd_open = self.get_command_object("cmdOpen")
            self.cmd_close = self.get_command_object("cmdClose")

            self.chan_state_open = self.get_channel_object("chanStateOpen")
            self.chan_state_closed = self.get_channel_object("chanStateClosed")

            if self.chan_state_open is not None:
                self.chan_state_open.connect_signal("update", self.state_open_changed)
            else:
                self.log.error("Cannot get channel for shutter %s" % self.username)

            if self.chan_state_closed is not None:
                self.chan_state_closed.connect_signal(
                    "update", self.state_closed_changed
                )
            else:
                self.log.error("Cannot get channel for shutter %s" % self.username)

            self.state_open_changed(self.chan_state_open.get_value())
        else:
            self.simulated_update()

        super().init()

    def _initialise_values(self):
        """Add additional, known in advance states to VALUES"""
        values_dict = {item.name: item.value for item in self.VALUES}
        values_dict.update({"MOVING": "MOVING"})
        self.VALUES = Enum("ValueEnum", values_dict)

    def get_value(self):
        if self.simulation:
            return self.simulated_update()

        if self._nominal_value == self.VALUES.MOVING:
            if (time.time() - self.cmd_started) < self.move_time_min:
                return self.VALUES.MOVING

        self.cover_is_open = self.chan_state_open.get_value()
        self.cover_is_closed = self.chan_state_closed.get_value()
        return self.update_cover_state()

    def _set_value(self, value):
        if self.simulation:
            if value == self.VALUES.OPEN:
                self.simulated_opened = 1
            else:
                self.simulated_opened = 0
            self.simulated_moving = True
            gevent.spawn(self.simul_do)
            return

        current_value = self.get_value()

        if current_value == self.VALUES.MOVING:
            self.log.error("Cannot move while moving")
            return

        if value != current_value:
            if value == self.VALUES.OPEN:
                self.cmd_open()
            elif value == self.VALUES.CLOSED:
                self.cmd_close()

            self.cmd_started = time.time()
            self.update_value(self.VALUES.MOVING)

    def simul_do(self):
        gevent.sleep(3)
        self.simulated_moving = False
        self.simulated_update()

    def is_moving(self):
        return self.get_value() == self.VALUES.MOVING

    def state_open_changed(self, value):
        """Updates cover state when cover open value changes

        :param state: cover open state
        :type state: str
        :return: None
        """
        self.cover_is_open = value
        self.update_cover_state()

    def state_closed_changed(self, value):
        """Updates cover state when cover close value changes

        :param state: cover close state
        :type state: str
        :return: None
        """
        self.cover_is_closed = value
        self.update_cover_state()

    def update_cover_state(self):
        """Updates cover state

        :return: cover state as str
        """
        if self.cover_is_open:
            value = self.VALUES.OPEN
        elif self.cover_is_closed:
            value = self.VALUES.CLOSED
        else:
            if time.time() - self.cmd_started > self.cmd_timeout:
                value = self.VALUES.UNKNOWN
            else:
                value = self.VALUES.MOVING

        self.update_value(value)

        return value

    def simulated_update(self):
        if self.simulated_moving:
            value = self.VALUES.MOVING
        elif self.simulated_opened:
            value = self.VALUES.OPEN
        else:
            value = self.VALUES.CLOSED

        self.update_value(value)
        return value

    @property
    def is_closed(self):
        """Check if the shutter is closed.
        Returns:
            (bool): True if open, False otherwise.
        """
        return self.get_value() == self.VALUES.CLOSED



import logging
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject


class ALBABackLight(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)
        self.limits = [None, None]
        self.state = None
        self.current_level = None
        self.actuator_status = None
        self.register_state = None

        self.memorized_level = None
        self.default_rest_level = 0.0
        self.default_minimum_level = 7.0

    def init(self):
        self.backlightin_channel = self.get_channel_object("backlightin")
        self.level_channel = self.get_channel_object("light_level")

        limits = self.get_property("limits")

        if limits is not None:
            lims = limits.split(",")
            if len(lims) == 2:
                self.limits = map(float, lims)

        rest_level = self.get_property("rest_level")

        if rest_level is not None:
            self.rest_level = rest_level
        else:
            self.rest_level = self.default_rest_level

        minimum_level = self.get_property("minimum_level")
        if minimum_level is not None:
            self.minimum_level = float(minimum_level)
        else:
            self.minimum_level = self.default_minimum_level

        self.level_channel.connect_signal("update", self.level_changed)
        self.backlightin_channel.connect_signal("update", self.state_changed)

    def is_ready(self):
        return True

    def level_changed(self, value):
        self.current_level = value
        self.emit("levelChanged", self.current_level)

    def state_changed(self, value):
        state = value
        if state != self.state:
            self.state = state
            if state:
                self.emit("stateChanged", "on")
            else:
                self.emit("stateChanged", "off")

    def _current_state(self):
        state = None

        if self.actuator_status:
            state = "off"
        else:
            state = "on"

        return state

    def get_limits(self):
        return self.limits

    def get_state(self):
        _state = self.backlightin_channel.get_value()
        if _state:
            return "on"
        else:
            return "off"

    def getUserName(self):
        return self.username

    def getLevel(self):
        self.current_level = self.level_channel.get_value()
        return self.current_level

    def setLevel(self, level):
        self.level_channel.set_value(float(level))

    def setOn(self):
        self.on_task = gevent.spawn(self._setOn)
        self.on_task.link(self._task_finished)
        self.on_task.link_exception(self._task_failed)

    def _setOn(self):
        if self.backlightin_channel.get_value() is False:
            self.set_backlight_in()
            wait_ok = self.wait_backlight_in()
            if not wait_ok:
                logging.getLogger("HWR").debug("could not set backlight in")
                return

        level = None
        if self.memorized_level:
            level = self.memorized_level

        if not level or level < self.minimum_level:
            level = self.minimum_level

        logging.getLogger("HWR").debug("setting light level to : %s" % level)
        self.setLevel(level)

    def set_backlight_in(self):
        self.backlightin_channel.set_value(True)

    def wait_backlight_in(self, state=True, timeout=10):
        t0 = time.time()
        elapsed = 0
        while elapsed < timeout:
            isin = self.backlightin_channel.get_value()
            if isin == state:
                logging.getLogger("HWR").debug(
                    "waiting for backlight took %s . In is: %s" % (elapsed, isin)
                )
                return True
            gevent.sleep(0.1)
            elapsed = time.time() - t0

        logging.getLogger("HWR").debug("Timeout waiting for backlight In")
        return False

    def _task_finished(self, g):
        logging.getLogger("HWR").debug("Backlight task finished")
        self._task = None

    def _task_failed(self, g):
        logging.getLogger("HWR").debug("Backlight task failed")
        self._task = None

    def setOff(self):
        if self.current_level:
            self.memorized_level = self.current_level
            self.setLevel(self.rest_level)
        self.backlightin_channel.set_value(False)


def test_hwo(hwo):
    import sys

    print('\nLight control for "%s"\n' % hwo.getUserName())
    print("   Level limits are:", hwo.get_limits())
    print("   Current level is:", hwo.getLevel())
    print("   Current state is:", hwo.get_state())
    print("   Setting backlight in")

    print(sys.argv)
    if sys.argv[3] == "0":
        print("Setting backlight off")
        n = False
        hwo.setOff()
    else:
        print("Setting backlight on")
        n = True
        hwo.setOn()

    hwo.wait_backlight_in(state=n)
    # while gevent.wait(timeout=0.1):
    #    print "Waiting"
    #    gevent.sleep(0.1)

    print("   Current state is:", hwo.get_state())



import json
import logging
import os
import time
from contextlib import ExitStack
from uuid import uuid1

import gevent
from gevent import subprocess
from lima2.client import Detector
from lima2.client.smx.aggregation.writer import Writer as SmxAggregationWriter

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.CommandContainer import ConnectionError
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector

_logger = logging.getLogger("HWR")

_logger_det = logging.getLogger("lima2.client.detector")
_logger_smx = logging.getLogger("lima2.client.smx")
_logger_smx_aggr = logging.getLogger("lima2.client.smx.aggregation")

_logger_det.setLevel(logging.DEBUG)
_logger_smx.setLevel(logging.DEBUG)
_logger_smx_aggr.setLevel(logging.DEBUG)

lima2_loggers = [_logger_det, _logger_smx, _logger_smx_aggr]


def update_lima2_loggers():
    for tgt_logger in lima2_loggers:
        tgt_logger.setLevel(_logger.getEffectiveLevel())
        for handler in _logger.handlers:
            tgt_logger.addHandler(handler)


# Logger decorator
def logger(fn):
    def inner(*args, **kwargs):
        _logger.debug(f"Entering %s", fn.__name__)
        to_execute = fn(*args, **kwargs)
        _logger.debug(f"Exiting %s", fn.__name__)
        return to_execute

    return inner


def convert_state(state):
    """Convert detector state to MxCube HWR state"""
    # UNKNOWN = 0
    # WARNING = 1
    # BUSY = 2
    # READY = 3
    # FAULT = 4
    # OFF = 5
    if state == Detector.State.IDLE or state == Detector.State.PREPARED:
        s = HardwareObjectState.READY
    elif state == Detector.State.RUNNING:
        s = HardwareObjectState.BUSY
    else:
        s = HardwareObjectState.UNKNOWN
    return s


def create_directory(path, check=True):
    subprocess.run(
        "mkdir --parents {0} && chmod -R 755 {0}".format(path),
        shell=True,
        check=check,
    )


class Lima2Detector(AbstractDetector):
    def __init__(self, name):
        AbstractDetector.__init__(self, name)
        self.header = {}
        self.start_angles = []
        self.__device = None

    def init(self):
        AbstractDetector.init(self)

        update_lima2_loggers()
        self.image_rejection_settings_file = self.get_property(
            "image_rejection_settings_file"
        )

        lima_ctrl_device = self.get_property("lima_ctrl_device", "")
        # lima_recv_devices = ast.literal_eval(self.get_property("lima_recv_devices", ""))
        lima_recv_devices = self.get_property("lima_recv_devices", "").split(",")

        _logger.info(
            "Initializing Lima2Detector: %s %s", lima_ctrl_device, lima_recv_devices
        )

        if not lima_ctrl_device or not len(lima_recv_devices) >= 1:
            return

        self.__proc_loop_task = None
        self.__proc_stalled = None
        self.__started = False
        self.__stopped = False

        try:
            self.__device = Detector(lima_ctrl_device, *lima_recv_devices)

            self.__acq_params = self.__device.acq_params
            self.__proc_params = self.__device.proc_params

            # Monitor device state
            def on_state_change(state):
                s = convert_state(state)
                _logger.info("State changed to %s / %s", state, s)
                self.update_state(s)

            self.__device.registerStateLogger(on_state_change)
        except (ConnectionError, AttributeError):
            self.update_state(HardwareObjectState.FAULT)
            _logger.error("Could not connect to detector %s" % lima_ctrl_device)
            self._emit_status()

    def has_shutterless(self):
        return True

    @logger
    def wait_idle(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not idle")):
            idle_states = [
                Detector.State.PREPARED,
                Detector.State.IDLE,
                Detector.State.FAULT,
            ]
            while self.__device.state not in idle_states:
                _logger.debug("State: %s", self.__device.state)
                gevent.sleep(1)
        gevent.sleep(2)
        proc, stalled = self.__proc_loop_task, self.__proc_stalled
        if proc:
            _logger.debug("waiting for processing to finish")
            gevent.wait([proc, stalled])
            if stalled.is_set():
                raise RuntimeError("Processing stalled")

    @logger
    def wait_prepared(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not prepared")):
            while self.__device.state != Detector.State.PREPARED:
                gevent.sleep(1)

    @logger
    def last_image_saved(self):
        try:
            img = 0  # TODO
            return img
        except Exception:
            return 0

    def get_deadtime(self):
        return float(self.get_property("deadtime"))

    def find_next_pedestal_dir(self, data_root_path, subdir):
        _index = 1
        _indes_str = "%04d" % _index
        fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        while os.path.exists(fpath):
            _index += 1
            _indes_str = "%04d" % _index
            fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        return fpath

    def set_detector_filenames(self, data_root_path, prefix):
        create_directory(data_root_path)

    @logger
    def prepare_acquisition(
        self, number_of_images, exptime, data_root_path, prefix, dense_skip_nohits=False
    ):
        self.__started = False
        self.__stopped = False

        if number_of_images < 1:
            msg = f"Lima2: Invalid number_of_images: {number_of_images}"
            _logger.error(msg)
            raise ValueError(msg)

        update_lima2_loggers()

        dump_params_file_pattern = "lima2_params-%Y-%m-%d-%H%M%S.json"
        file_name = time.strftime(dump_params_file_pattern)
        dump_params_filename = os.path.join(data_root_path, file_name)

        packet_fifo_depth = 20000

        acq_params = {
            "acq": {
                "expo_time": int(exptime * 1e6),
                # "latency_time": 990,
                "nb_frames": number_of_images,
                "trigger_mode": "external",
                "nb_frames_per_trigger": 1,
            },
            "det": {
                "gain_mode": "dynamic",
                "packet_fifo_depth": packet_fifo_depth,
                "auto_comparator_disable": True,
            },
        }

        self.set_detector_filenames(data_root_path, prefix)

        saving_groups = ["raw", "dense", "sparse", "spots", "accumulated"]

        # all streams will be saved in dedicated sub dirs except dense
        data_sub_dir = {g: g if g != "dense" else "" for g in saving_groups}

        data_path = {
            g: os.path.join(data_root_path, s) if s else data_root_path
            for g, s in data_sub_dir.items()
        }

        acc_frames = 1000

        max_nb_frames_per_file = {
            "raw": 1000,
            "dense": 1000,
            "sparse": 1000,
            "spots": 1000,
            "accumulated": 1,
        }

        nb_recvs = len(self.__device.recvs)
        nb_recv_frames = number_of_images // nb_recvs
        nb_acc_frames = (number_of_images - 1) // acc_frames + 1

        nb_frames = {
            g: nb_acc_frames if g == "accumulated" else nb_recv_frames
            for g in saving_groups
        }

        def calc_frame_per_file(group, n):
            return min(max_nb_frames_per_file[group], n)

        frames_per_file = {
            g: calc_frame_per_file(g, nb_frames[g]) for g in saving_groups
        }

        save_files = {
            "raw": False,
            "dense": True,
            "sparse": True,
            "spots": True,
            "accumulated": True,
        }

        dense_comp_with_hw_nx = True
        dense_comp = "zip" if dense_comp_with_hw_nx else "bshuf_lz4"

        compression = {
            "raw": dense_comp,
            "dense": dense_comp,
            "sparse": "none",
            "spots": "none",
            "accumulated": "zip",
        }

        saving_streams = ["raw", "dense", "sparse", "spots"] + [
            f"accumulation_{a}" for a in ["corrected", "ishit", "nohit"]
        ]

        def get_stream_group(stream):
            is_acc = stream.startswith("accumulation_")
            return "accumulated" if is_acc else stream

        def get_stream_prefix(stream):
            return stream.replace("accumulation", "acc")

        def get_saving(stream):
            group = get_stream_group(stream)
            stream_prefix = get_stream_prefix(stream)
            return {
                "enabled": save_files[group],
                "base_path": data_path[group],
                "filename_prefix": f"{prefix}_{stream_prefix}",
                "start_number": 0,
                "nb_frames_per_file": frames_per_file[group],
                "file_exists_policy": "abort",
                "compression": compression[group],
            }

        sub_dirs = [s for g, s in data_sub_dir.items() if s and save_files[g]]
        # the sub directory for aggregated data from all receivers
        sub_dirs.append("aggregated")
        for s in sub_dirs:
            create_directory(os.path.join(data_root_path, s))

        fai_kernels_base = self.get_property("fai_kernels_base", ".")
        params_base = self.get_property("params_base", ".")
        pedestal_path = os.path.join(data_root_path, "pedestal.h5")

        config_beam = None
        config_beam_dict = self.get_property(f"beam", {})
        if config_beam_dict:
            _logger.debug("config_beam_dict=%s", config_beam_dict)
            config_beam = [float(config_beam_dict.get(f"b{i}", "0.0")) for i in "xy"]

        mask_beam_stop = True
        mask_filename = "mask.h5" if mask_beam_stop else "mask_no_beamstop.h5"

        manage_proc = True

        # energy = 11.56
        energy = HWR.beamline.energy.get_value()
        # energy = 12.75
        jungfrau_gain0_ave = 41.401

        def get_param_file(n):
            return os.path.join(params_base, n)

        fai_params = {
            "mask_path": get_param_file(mask_filename),
            "csr_path": get_param_file("csr.h5"),
            "radius2d_path": get_param_file("r_center.h5"),
            "radius1d_path": get_param_file("bin_centers.h5"),
            "error_model": "azimuthal",
            "dummy": 0.0,
            "delta_dummy": 0.0,
            "normalization_factor": 1.0,
            "cutoff_clip": 0,
            "cycle": 3,
            # "empty": -9999.0,
            "noise": 0.5,
            "cutoff_pick": 4.0,
            "patch_size": 5,
            "connected": 7,
            "min_nb_peaks": 20,
            "dense_skip_nohits": dense_skip_nohits,
            "acc_nb_frames_reset": 0,
            "acc_nb_frames_xfer": acc_frames,
        }

        legacy_photon_adus = jungfrau_gain0_ave * energy

        def get_dense_out_params(variant_name):
            variant_data = {
                "i32": {
                    "pixel_type": "int32",
                    "photon_adus": legacy_photon_adus,
                    "photon_bias": 0.0,
                },
                "i16_1": {
                    "pixel_type": "int16",
                    "photon_adus": 16.0,
                    "photon_bias": 0.0,
                },
                "i16_2": {
                    "pixel_type": "int16",
                    "photon_adus": 8.0,
                    "photon_bias": 0.0,
                },
                "i16_3": {
                    "pixel_type": "int16",
                    "photon_adus": 1.0,
                    "photon_bias": 0.0,
                },
                "u16_1": {
                    "pixel_type": "uint16",
                    "photon_adus": 16.0,
                    "photon_bias": 32.0,
                },
                "u16_2": {
                    "pixel_type": "uint16",
                    "photon_adus": 8.0,
                    "photon_bias": 32.0,
                },
                "f16": {
                    "pixel_type": "float16",
                    "photon_adus": 1.0,
                    "photon_bias": 0.0,
                },
            }

            variant = variant_data[variant_name]
            photon_adus = variant["photon_adus"]
            return {
                "dense_intensity_factor": photon_adus,
                "dense_intensity_offset": photon_adus * variant["photon_bias"],
                "dense_pixel_type": "dense_%s" % variant["pixel_type"],
            }

        dense_variant_name = "i32"
        dense_out_params = get_dense_out_params(dense_variant_name)
        fai_params.update(dense_out_params)

        # pf_params_fname = get_param_file("peakfinder_params.json")
        # if os.path.exists(pf_params_fname):
        #     with open(pf_params_fname, "rt") as f:
        #         pf_params = json.load(f)

        fpath = HWR.get_hardware_repository().find_in_repository(
            self.image_rejection_settings_file
        )

        if os.path.exists(fpath):
            _logger.info(f"Reading parameters from {fpath}")
            with open(fpath, "rt") as f:
                pf_params = json.load(f)

            _logger.info(f"Parameters: {pf_params}")

        pixel_size = 75e-6

        _logger.info("peakfinder params: %s", pf_params)
        # the "threshold" parameter will always refer to int32 scale
        radius_max = pf_params.get("radius_max", 0.0) * pixel_size
        fai_user_params = {
            "radius_max": radius_max,
            "cutoff_pick": pf_params["snr"],
            "noise": pf_params["threshold"] / legacy_photon_adus,
            "patch_size": pf_params["patch_size"],
            "connected": pf_params["connected"],
            "min_nb_peaks": pf_params["min_nb_peaks"],
            "dense_skip_nohits": dense_skip_nohits,
            # "dense_skip_nohits": pf_params["discard_no_hits"],
        }
        fai_params.update(fai_user_params)

        proc_params = {
            "proc_mode": "fai",
            "fifo": {
                "nb_fifo_frames": packet_fifo_depth,
            },
            "buffers": {
                "nb_peak_counters_buffer": nb_recv_frames,
            },
            "gpu": {
                "device_idx": 0,
                "cl_source_path": fai_kernels_base,
            },
            "jfrau": {
                "gain_path": get_param_file("gains.h5"),
                "pedestal_path": pedestal_path,
                "photon_energy": energy,
            },
            "fai": fai_params,
        }

        _logger.info("FAI PARAMS: %s", fai_params)
        _logger.info("PARAMS: %s", proc_params)

        proc_params.update(
            {f"saving_{stream}": get_saving(stream) for stream in saving_streams}
        )

        self.__device.nb_gpus_per_system = 2

        self.__device.acq_params.update(acq_params)
        _logger.debug("acq_params: %s", self.__device.acq_params)
        self.__device.proc_params.update(proc_params)
        _logger.debug("proc_params: %s", self.__device.proc_params)

        uuid = uuid1()
        _logger.info(f"UUID={uuid}")

        lima2_params = self.__device.prepareAcq(uuid)
        if dump_params_filename:
            _logger.info("saving lima2 params to %s", dump_params_filename)
            with open(dump_params_filename, "wt") as f:
                json.dump(lima2_params, f, indent=4, sort_keys=True)

        # Async version
        # gevent.spawn(self.__device.prepareAcq, uuid).link_value(on_prepared)

        if self.__stopped:
            return

        # Read calibration data
        calib_params_file = get_param_file("calib_params.txt")
        with open(calib_params_file, "rt") as f:
            calib_params = json.loads(f.read())
        _logger.debug("calib_params: %s", calib_params)

        sample_distance = calib_params["sample_distance"]
        beam_center = calib_params["beam_center"]
        if config_beam and all(config_beam):
            if any([fabs(c - b) > 1 for c, b in zip(config_beam, beam_center)]):
                _logger.warning(
                    "config beam (%s) differs from beam_center (%s)",
                    config_beam,
                    beam_center,
                )

        # Master file header: metadata
        hc_over_e = 12.398419
        wavelength = hc_over_e / energy
        adus_per_photon = dense_out_params["dense_intensity_factor"]
        bias_adus = dense_out_params["dense_intensity_offset"]
        photon_dynamic_range = [-1.0, 1e4]
        trusted_range = [
            int(round(p * adus_per_photon + bias_adus)) for p in photon_dynamic_range
        ]

        header = {
            "beam": {"incident_wavelength": (wavelength, "angstrom")},
            "detector_information": {
                "detector_distance": (sample_distance * 1e-3, "m"),
                "detector_name": "jungfrau-4m",
                "x_pixel_size": (pixel_size, "m"),
                "y_pixel_size": (pixel_size, "m"),
                "beam_center_x": (beam_center[0], "pixel"),
                "beam_center_y": (beam_center[1], "pixel"),
                "adus_per_photon": adus_per_photon,
                "underload_value": trusted_range[0],
                "saturation_value": trusted_range[1],
                "trusted_range": trusted_range,
            },
        }

        # Manage processing
        if manage_proc:
            self.__proc_stalled = gevent.event.Event()
            self.__proc_loop_task = gevent.spawn(self.proc_loop, uuid, header)

    @logger
    def start_acquisition(self):
        self.wait_prepared()
        if not self.__stopped:
            self.__device.startAcq()
            self.__started = True

    @logger
    def stop_acquisition(self):
        self.__stopped = True
        wait_for_idle_timeout = 10
        try:
            self.wait_idle(wait_for_idle_timeout)
        except:
            _logger.warning("detector not idle after %s", wait_for_idle_timeout)

        try:
            if self.__started:
                self.__device.stopAcq()
        except Exception:
            if self.__proc_loop_task:
                gevent.kill(self.__proc_loop_task)
            self.__device.resetAcq()
        finally:
            self.wait_idle()

    @logger
    def proc_loop(self, uuid, header={}):
        with ExitStack() as stack:

            def cleanup():
                self.__proc_loop_task = None
                self.__proc_stalled = None

            stack.callback(cleanup)

            _logger.debug("starting processing loop for %s", uuid)
            proc_cm = self.__device.getProcessing(uuid, erase_on_cleanup=True)
            proc = stack.enter_context(proc_cm)
            stack.callback(_logger.debug, "erasing processing for %s", uuid)

            # Aggregation (master file) writer
            det_name = proc.detector.det_name  # or self.name()
            local_2_shared_dir_map = {
                os.path.join(
                    "/nobackup/lid29p9jfrau11/shared",
                    "lima2/detectors/psi/data/processing",
                    "jungfrau_4m_01",
                ): os.path.join("/data/id29/inhouse/opid291/Jungfrau", "Calibration"),
            }
            writer_args = {
                "det_name": det_name,
                "master_subdir": "aggregated",
                "header": header,
                "local_2_shared_dir_map": local_2_shared_dir_map,
            }
            stack.enter_context(SmxAggregationWriter(proc, **writer_args))

            def last_saved():
                cnts = proc.progress_counters
                names = ["dense_saved", "sparse_saved", "spots_saved"]
                return max([getattr(cnts, f"nb_frames_{n}") for n in names])

            stalled_timeout = 10
            last_cnt = 0
            last_t = None

            # main processing loop
            while not proc.is_finished:
                if not self.__started and self.__stopped:
                    break

                # Check if processing is stalled
                l = last_saved()
                if l > last_cnt:
                    last_cnt = l
                    last_t = time.time()
                if last_t and time.time() - last_t > stalled_timeout:
                    _logger.debug("lima2 processing stalled")
                    self.__proc_stalled.set()

                gevent.sleep(1.0)

    def get_acquired_frames(self):
        return self.__device.nb_frames_xferred

    @logger
    def reset(self):
        self.__device.resetAcq()

    @property
    def lima2_device(self):
        return self.__device

    @property
    def status(self):
        try:
            acq_status = self.__device.state
        except Exception:
            acq_status = "OFFLINE"

        status = {
            "acq_satus": str(acq_status).upper(),
        }

        return status

    def _emit_status(self):
        self.emit("statusChanged", self.status)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import copy
import logging

from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)
from mxcubecore.dispatcher import saferef

from .. import Poller
from .. import TacoDevice_MTSafe as TacoDevice

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class TacoCommand(CommandObject):
    def __init__(
        self, name, command, taconame=None, username=None, args=None, dc=False, **kwargs
    ):
        CommandObject.__init__(self, name, username, **kwargs)

        self.command = command
        self.device_name = taconame
        self.data_collector = dc
        self.pollers = {}
        self.__value_changed_callback_ref = None
        self.__timeout_callback_ref = None

        if args is None:
            self.arg_list = ()
        else:
            # not very nice...
            args = str(args)
            if not args.endswith(","):
                args += ","
            self.arg_list = eval("(" + args + ")")

        self.connect_device()

    def connect_device(self):
        try:
            self.device = TacoDevice.TacoDevice(
                self.device_name, dc=self.data_collector
            )
        except Exception:
            logging.getLogger("HWR").exception(
                "Problem with Taco ; could not open Device %s", self.device_name
            )
            self.device = None

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))

        if len(args) > 0 and len(self.arg_list) > 0:
            logging.getLogger("HWR").error(
                "%s: cannot execute command with arguments when 'args' is defined from XML",
                str(self.name()),
            )
            self.emit("commandFailed", (-1, str(self.name())))
            return
        elif len(args) == 0 and len(self.arg_list) > 0:
            args = self.arg_list

        if self.device is not None and self.device.imported:
            try:
                ret = eval("self.device.%s(*%s)" % (self.command, args))
            except Exception:
                logging.getLogger("HWR").error(
                    "%s: an error occured when calling Taco command %s",
                    str(self.name()),
                    self.command,
                )
            else:
                self.emit("commandReplyArrived", (ret, str(self.name())))
                return ret

        self.emit("commandFailed", (-1, str(self.name())))

    def _value_changed(self, value):
        self.value_changed(self.device_name, value)

    def value_changed(self, device_name, value):
        try:
            callback = self.__value_changed_callback_ref()
        except Exception:
            pass
        else:
            if callback is not None:
                callback(device_name, value)

    def on_polling_error(self, exception, poller_id):
        self.connect_device()
        poller = Poller.get_poller(poller_id)
        if poller is not None:
            try:
                poller.restart(1000)
            except Exception:
                pass

    def poll(
        self,
        polling_time=500,
        arguments_list=(),
        value_changed_callback=None,
        timeout_callback=None,
        direct=True,
        compare=True,
    ):
        self.__value_changed_callback_ref = saferef.safe_ref(value_changed_callback)

        poll_cmd = getattr(self.device, self.command)

        Poller.poll(
            poll_cmd,
            copy.deepcopy(arguments_list),
            polling_time,
            self._value_changed,
            self.on_polling_error,
            compare,
        )

    def stopPolling(self):
        pass

    def abort(self):
        pass

    def is_connected(self):
        return self.device is not None and self.device.imported


class TacoChannel(ChannelObject):
    """Emulation of a 'Taco channel' = a Taco command + polling"""

    def __init__(
        self,
        name,
        command,
        taconame=None,
        username=None,
        polling=None,
        args=None,
        **kwargs,
    ):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.command = TacoCommand(
            name + "_internalCmd", command, taconame, username, args, False, **kwargs
        )

        try:
            self.polling = int(polling)
        except Exception:
            self.polling = None
        else:
            self.command.poll(self.polling, self.command.arg_list, self.value_changed)

    def value_changed(self, deviceName, value):
        self.emit("update", value)

    def get_value(self):
        return self.command()

    def is_connected(self):
        return self.command.is_connected()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import ast
import time

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.NState import NState


class P11BeamStop(NState):
    default_delta = 0.1

    def init(self):
        """Initialize the BeamStop motors and load positions."""
        super().init()

        self.x_motor = self.get_object_by_role("bstopx")

        self.load_positions()
        self.load_deltas()

        self.log.debug(f"Beamstop X Motor initialized: {self.x_motor}")
        self.log.debug(f"Beamstop Y Motor not used")
        self.log.debug(f"Beamstop Z Motor not used")

    def load_positions(self):
        """Load predefined positions from the XML configuration."""
        self.log.info("Loading BeamStop positions from config")
        positions_str = self.get_property("values")

        if not positions_str:
            self.log.error(
                "No values for BeamStop positions found in the configuration."
            )
            raise ValueError("No BeamStop positions found in configuration")

        # Convert the string to a dictionary using ast.literal_eval
        try:
            self.positions = ast.literal_eval(positions_str)
            if isinstance(self.positions, dict):
                self.log.info(f"Available BeamStop positions: {self.positions}")
            else:
                raise ValueError("Positions data is not a dictionary")
        except (SyntaxError, ValueError) as e:
            self.log.error(f"Error parsing BeamStop positions: {e}")
            raise ValueError("Invalid BeamStop positions format in the configuration.")

    def load_deltas(self):
        """Load individual motor deltas from the XML configuration explicitly."""
        self.log.info("Loading deltas from config")

        # Fetch individual deltas for each motor
        delta_bstopx = self.get_property("delta_bstopx")

        # If a delta is not specified, fallback to a default delta value
        self.deltas = {
            "bstopx": (
                float(delta_bstopx) if delta_bstopx is not None else self.default_delta
            ),
        }

        # Log the deltas for each motor
        for motorname, delta in self.deltas.items():
            self.log.info(f"Delta for {motorname}: {delta}")

    def set_value(self, value):
        """Move the BeamStop motors to the given position."""
        if value not in self.positions:
            raise ValueError(f"Invalid value {value}, not in available positions")

        position = self.positions[value]

        self.x_motor._set_value(position.get("bstopx"))

    def get_value(self):
        """Get the current BeamStop position based on the motor positions."""
        current_x = self.x_motor.get_value()

        for position_name, position in self.positions.items():
            if self.is_within_deltas(position.get("bstopx"), current_x, "bstopx"):
                return position_name  # Return the matching position name

    def is_within_deltas(self, target_value, current_value, motor_name):
        """Check if the current motor position is within the delta tolerance for that specific motor."""
        delta = self.deltas.get(motor_name)
        if target_value is None or delta is None:
            return False
        return abs(current_value - target_value) <= delta

    def get_position_list(self):
        """Return a list of available positions."""
        return list(self.positions.keys())

    def wait_for_position(self):
        """Wait for motors to reach their target positions."""
        while any(motor.get_state() != "ON" for motor in [self.x_motor]):
            time.sleep(0.1)

    def is_moving(self):
        """
        Descript. : True if the motor is currently moving
        """
        return self.get_state() == HardwareObjectState.BUSY



"""
[Name] MachineInfoMockup

[Description]
MachineInfo hardware objects are used to obtain information from the
accelerator control system.

This is a mockup hardware object, it simulates the behaviour of an accelerator
information by :

    - produces a current value that varies with time
    - simulates a control room message that changes with some condition
      ()
    - simulates

[Emitted signals]
valueChanged
   pars:  values (dict)

   mandatory fields:
     values['current']  type: str; desc: synchrotron radiation current in milli-amps
     values['message']  type: str; desc: message from control room
     values['attention'] type: boolean; desc: False (no attention required)
                                            True (attention raised to the user)

   optional fields:
      any number of optional fields can be sent over with this signal by
      adding them in the values dictionary

      for example:
         values['lifetime']
         values['topup_remaining']
"""

import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractMachineInfo import AbstractMachineInfo


class MachineInfoMockup(AbstractMachineInfo):
    """Simulates the behaviour of an accelerator information"""

    default_current = 200  # [mA]
    default_message = "Beam Delivered"
    default_lifetime = 45  # hours Lifetime
    default_topup_remaining = 70  # [s]

    def init(self):
        """Initialise some parameters and update routine."""
        super().init()
        self._current = self.default_current
        self._message = self.default_message
        self._lifetime = self.default_lifetime
        self._topup_remaining = self.default_topup_remaining
        self._run()

    def _run(self):
        """Spawn update routine."""
        gevent.spawn(self._update_me)

    def _update_me(self):
        """Simulate change of different parameters"""
        self.t0 = time.time()

        while True:
            gevent.sleep(5)
            elapsed = time.time() - self.t0
            self._topup_remaining = abs((self.default_topup_remaining - elapsed) % 300)
            if self._topup_remaining < 60:
                self._message = f"ATTENTION: topup in {self._topup_remaining} s"
                self.attention = True
            else:
                self._message = self.default_message
                self.attention = False

            self._current = f"{(self.default_current - (3 - self._topup_remaining / 100.0) * 5):3.2f}"

            values = {}
            values["message"] = self._message
            values["topup_remaining"] = self._topup_remaining
            values["attention"] = self.attention

            # current and lifetime should be configured in the xml file
            values.update(self.get_value())
            self.update_value(values)

    def get_current(self) -> float:
        """Override method."""
        return self._current

    def get_lifetime(self) -> float:
        """Override method."""
        return self._lifetime

    def get_topup_remaining(self) -> float:
        """Override method."""
        return self._topup_remaining

    def get_message(self) -> str:
        """Override method."""
        return self._message


def test():
    """Test routine"""
    import sys

    hwr = HWR.get_hardware_repository()
    hwr.connect()

    conn = hwr.get_hardware_object(sys.argv[1])

    print(("Machine current: ", conn.get_current()))
    print(("Life time: ", conn.get_lifetime()))
    print(("TopUp remaining: ", conn.get_topup_remaining()))
    print(("Message: ", conn.get_message()))
    print(("Values: ", conn.get_mach_info_dict()))

    while True:
        gevent.wait(timeout=0.1)


if __name__ == "__main__":
    test()



import logging

import gevent
from pydantic.v1 import (
    Field,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SSXLaserHareCollectionParameters(BaseUserCollectionParameters):
    align_chip: bool = Field(True)

    class Config:
        extra: "ignore"


class SsxLaserHareColletionTaskParameters(SsxBaseQueueTaskParameters):
    user_collection_parameters: SSXLaserHareCollectionParameters


class SsxLaserHareCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxLaserHareCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxLaserHareCollectionQueueModel
    DATA_MODEL = SsxLaserHareColletionTaskParameters
    NAME = "SSX Laser HARE Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    def __init__(self, view, data_model: SsxLaserHareCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)
        self._data_model._task_data.collection_parameters.num_images = 400

        self.__stop_req = False
        self.__stop_done = True

    def execute(self):
        super().execute()

        self.__stop_req = False
        self.__stop_done = False

        params = self._data_model._task_data.user_collection_parameters

        data_root_path = self.get_data_path()
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        enforce_centring_phase = False
        reject_empty_frames = (
            self._data_model._task_data.user_collection_parameters.reject_empty_frames
        )
        delay = HWR.beamline.diffractometer.get_ssx_delay()
        ssx_laser_scan_method = HWR.beamline.diffractometer.get_ssx_scan_method()

        self.take_pedestal()

        self.start_processing("CHIP")

        selected_regions = self._data_model._task_data.collection_parameters.selection
        selected_regions = selected_regions if selected_regions else [[0, 0]]

        for region in selected_regions:
            if self.__stop_req:
                self.__stop_done = True
                logging.getLogger("user_level_log").info("Stopped sequence")
                break

            fname_prefix = self._data_model._task_data.path_parameters.prefix
            fname_prefix += f"_block_{region[0]}_{region[1]}_"

            # HWR.beamline.diffractometer.set_phase("Centring", wait=True, timeout=120)
            logging.getLogger("user_level_log").info(
                f"Laser scan method {ssx_laser_scan_method}"
            )
            logging.getLogger("user_level_log").info(f"Laser delay {delay}")
            logging.getLogger("user_level_log").info(f"Acquiring {region} ...")

            if params.align_chip:
                logging.getLogger("user_level_log").info(f"Aligning block {region}")
                HWR.beamline.diffractometer.set_phase(
                    "Centring", wait=True, timeout=120
                )
                HWR.beamline.diffractometer.prepare_ssx_grid_scan(
                    params.sub_sampling, region[0], region[1], False
                )
                logging.getLogger("user_level_log").info(f"Aligned block {region}")
                enforce_centring_phase = True

            logging.getLogger("user_level_log").info("Preparing detector")

            HWR.beamline.detector.stop_acquisition()
            gevent.sleep(10)
            HWR.beamline.detector.prepare_acquisition(
                400,
                params.exp_time,
                data_root_path,
                fname_prefix,
                dense_skip_nohits=reject_empty_frames,
            )

            HWR.beamline.diffractometer.wait_ready()

            logging.getLogger("user_level_log").info(f"Preparing data collection")
            HWR.beamline.diffractometer.set_phase("DataCollection")

            if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
                logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
                HWR.beamline.control.safshut_oh2.open()

            HWR.beamline.diffractometer.wait_ready()
            HWR.beamline.detector.wait_ready()
            gevent.sleep(5)

            HWR.beamline.detector.start_acquisition()
            logging.getLogger("user_level_log").info(
                "Detector ready, waiting for trigger ..."
            )

            try:
                logging.getLogger("user_level_log").info(f"Scanning {region} ...")
                HWR.beamline.diffractometer.start_ssx_scan(enforce_centring_phase)
            except:
                err_msg = "Diffractometer scan failed ..."
                logging.getLogger("user_level_log").exception(err_msg)
                HWR.beamline.detector.stop_acquisition()
                raise

            try:
                HWR.beamline.diffractometer.wait_ready()
            finally:
                HWR.beamline.detector.wait_ready()

            logging.getLogger("user_level_log").info(f"Acquired {region}")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        self.__stop_req = True
        super().stop()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Concrete Implementation of AbstartNState - overloads the get_value
and _set_value abstract methods.
Example xml configuration:

.. code-block:: xml

 <object class="NState">
   <username>Beam Monitoring</username>
   <actuator_name>beam_monitor</actuator_name>
   <values>{"ENABLED": True, "DISABLED": False}</values>
 </object>
"""

from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class NState(AbstractNState):
    """Overload get_value and _set_value abstract methods"""

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): VALUES Enum member
        """
        return self._nominal_value

    def _set_value(self, value):
        """Set the value.
        Args:
            value (Enum): target value
        """
        self.update_state(self.STATES.BUSY)
        self.update_value(value)
        self.update_state(self.STATES.READY)



import logging
import os
import traceback
from collections import namedtuple

import urllib2
from cookielib import CookieJar
from ISPyBClient import (
    _CONNECTION_ERROR_MSG,
    ISPyBClient,
)
from suds.client import Client
from suds.transport.http import HttpAuthenticated

from mxcubecore import HardwareRepository as HWR

# The WSDL root is configured in the hardware object XML file.
# _WS_USERNAME, _WS_PASSWORD have to be configured in the HardwareObject XML file.
_WSDL_ROOT = ""
_WS_BL_SAMPLE_URL = _WSDL_ROOT + "ToolsForBLSampleWebService?wsdl"
_WS_SHIPPING_URL = _WSDL_ROOT + "ToolsForShippingWebService?wsdl"
_WS_COLLECTION_URL = _WSDL_ROOT + "ToolsForCollectionWebService?wsdl"
_WS_AUTOPROC_URL = _WSDL_ROOT + "ToolsForAutoprocessingWebService?wsdl"
_WS_USERNAME = None
_WS_PASSWORD = None

SampleReference = namedtuple(
    "SampleReference",
    ["code", "container_reference", "sample_reference", "container_code"],
)


class SOLEILISPyBClient(ISPyBClient):
    def __init__(self, name):
        ISPyBClient.__init__(self, name)

        logger = logging.getLogger("ispyb_client")
        print("ISPYB")

        try:
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            hdlr = logging.FileHandler(
                "/home/experiences/proxima2a/com-proxima2a/MXCuBE_v2_logs/ispyb_client.log"
            )
            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)
        except Exception:
            pass

        logger.setLevel(logging.INFO)

    def init(self):
        """
        Init method declared by HardwareObject.
        """
        self.authServerType = self.get_property("authServerType") or "ldap"
        if self.authServerType == "ldap":
            # Initialize ldap
            self.ldapConnection = self.get_object_by_role("ldapServer")
            if self.ldapConnection is None:
                logging.getLogger("HWR").debug("LDAP Server is not available")

        self.loginType = self.get_property("loginType") or "proposal"
        self.loginTranslate = self.get_property("loginTranslate") or True
        self.beamline_name = HWR.beamline.session.beamline_name
        print("self.beamline_name init", self.beamline_name)

        self.ws_root = self.get_property("ws_root")
        self.ws_username = self.get_property("ws_username")
        if not self.ws_username:
            self.ws_username = _WS_USERNAME
        self.ws_password = self.get_property("ws_password")
        if not self.ws_password:
            self.ws_password = _WS_PASSWORD

        self.ws_collection = self.get_property("ws_collection")
        self.ws_shipping = self.get_property("ws_shipping")
        self.ws_tools = self.get_property("ws_tools")

        logging.info("SOLEILISPyBClient: Initializing SOLEIL ISPyB Client")
        logging.info("   - using http_proxy = %s " % os.environ["http_proxy"])

        try:
            if self.ws_root:
                logging.info(
                    "SOLEILISPyBClient: attempting to connect to %s" % self.ws_root
                )
                print("SOLEILISPyBClient: attempting to connect to %s" % self.ws_root)

                try:
                    self._shipping = self._wsdl_shipping_client()
                    self._collection = self._wsdl_collection_client()
                    self._tools_ws = self._wsdl_tools_client()
                    logging.debug(
                        "SOLEILISPyBClient: extracted from ISPyB values for shipping, collection and tools"
                    )
                except Exception:
                    print(traceback.print_exc())
                    logging.exception("SOLEILISPyBClient: %s" % _CONNECTION_ERROR_MSG)
                    return
        except Exception:
            print(traceback.print_exc())
            logging.getLogger("HWR").exception(_CONNECTION_ERROR_MSG)
            return
        try:
            proposals = HWR.beamline.session["proposals"]

            for proposal in proposals:
                code = proposal.code
                self._translations[code] = {}
                try:
                    self._translations[code]["ldap"] = proposal.ldap
                except AttributeError:
                    pass
                try:
                    self._translations[code]["ispyb"] = proposal.ispyb
                except AttributeError:
                    pass
                try:
                    self._translations[code]["gui"] = proposal.gui
                except AttributeError:
                    pass
        except IndexError:
            pass
        except Exception:
            pass

    def translate(self, code, what):
        """
        Given a proposal code, returns the correct code to use in the GUI,
        or what to send to LDAP, user office database, or the ISPyB database.
        """
        if what == "ispyb":
            return "mx"
        return ""

    def _wsdl_shipping_client(self):
        return self._wsdl_client(self.ws_shipping)

    def _wsdl_tools_client(self):
        return self._wsdl_client(self.ws_tools)

    def _wsdl_collection_client(self):
        return self._wsdl_client(self.ws_collection)

    def _wsdl_client(self, service_name):
        # Handling of redirection at soleil needs cookie handling
        cj = CookieJar()
        url_opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cj))

        trans = HttpAuthenticated(username=self.ws_username, password=self.ws_password)
        logging.info("_wsdl_client service_name %s - trans %s" % (service_name, trans))
        print("_wsdl_client service_name %s - trans %s" % (service_name, trans))

        trans.urlopener = url_opener
        urlbase = service_name + "?wsdl"
        locbase = service_name

        ws_root = self.ws_root.strip()

        url = ws_root + urlbase
        loc = ws_root + locbase

        print("_wsdl_client, url", url)
        ws_client = Client(url, transport=trans, timeout=3, location=loc, cache=None)

        return ws_client

    def prepare_collect_for_lims(self, mx_collect_dict):
        # Attention! directory passed by reference. modified in place

        prop = "EDNA_files_dir"
        path = mx_collect_dict[prop]
        ispyb_path = HWR.beamline.session.path_to_ispyb(path)
        mx_collect_dict[prop] = ispyb_path

        prop = "process_directory"
        path = mx_collect_dict["fileinfo"][prop]
        ispyb_path = HWR.beamline.session.path_to_ispyb(path)
        mx_collect_dict["fileinfo"][prop] = ispyb_path

        for i in range(4):
            try:
                prop = "xtalSnapshotFullPath%d" % (i + 1)
                path = mx_collect_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                logging.debug("SOLEIL ISPyBClient - %s is %s " % (prop, ispyb_path))
                mx_collect_dict[prop] = ispyb_path
            except Exception:
                pass

    def prepare_image_for_lims(self, image_dict):
        for prop in ["jpegThumbnailFileFullPath", "jpegFileFullPath"]:
            try:
                path = image_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                image_dict[prop] = ispyb_path
            except Exception:
                pass


def test_hwo(hwo):
    proposal_code = "mx"
    proposal_number = "20100023"
    proposal_psd = "tisabet"

    # proposal_number = '20160745'
    # proposal_psd = '087D2P3252'

    print("Trying to login to ispyb")
    info = hwo.login(proposal_number, proposal_psd)
    print("logging in returns: ", str(info))


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    db = HWR.beamline.lims

    print("db", db)
    print("dir(db)", dir(db))
    # print 'db._SOLEILISPyBClientShipping', db._SOLEILISPyBClientShipping
    # print 'db.Shipping', db.Shipping

    proposal_code = "mx"
    proposal_number = "20100023"
    proposal_psd = "tisabet"

    info = db.get_proposal(proposal_code, proposal_number)  # proposal_number)
    print(info)

    info = db.login(proposal_number, proposal_psd)
    print(info)


if __name__ == "__main__":
    test()



import logging
from typing import List

from mxcubecore.HardwareObjects.abstract.AbstractLims import AbstractLims
from mxcubecore.model.lims_session import (
    Lims,
    LimsSessionManager,
    Session,
)

logger = logging.getLogger("HWR")


class ESRFLIMS(AbstractLims):
    """
    ESRF client (ICAT+ and IPyB).
    """

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.drac = self.get_object_by_role("drac")
        self.ispyb = self.get_object_by_role("ispyb")

        self.is_local_host = False
        self.active_lims = self.drac.get_lims_name()[0]

    def get_lims_name(self) -> List[Lims]:
        return self.drac.get_lims_name() + self.ispyb.get_lims_name()

    def get_session_id(self) -> str:
        logger.debug("Setting up drac session_id=%s" % (self.drac.get_session_id()))
        return self.drac.get_session_id()

    def is_single_session_available(self):
        """
        True if there is no active session and there is
        a single session available
        """
        return (
            self.session_manager.active_session is None
            and len(self.session_manager.sessions) == 1
        )

    def login(self, user_name, token, is_local_host=False) -> LimsSessionManager:
        self.is_local_host = is_local_host
        session_manager, lims_username, sessions = self.drac.login(
            user_name, token, self.session_manager
        )
        logger.debug("%s sessions found. user=%s" % (len(sessions), user_name))

        self.session_manager = self.drac.session_manager

        self.add_user_and_shared_sessions(lims_username, sessions)

        # In case there is a single available session then it is selected automatically
        if self.is_single_session_available():
            single_session = self.session_manager.sessions[0]
            logger.debug(
                "Single session available which will be selected automatically. session_id=%s"
                % (single_session.session_id)
            )
            self.set_active_session_by_id(single_session.session_id)

        if session_manager.active_session is None:
            logger.debug(
                "DRAC no session selected then no activation of session in ISPyB"
            )
        else:
            self.ispyb.get_session_manager_by_code_number(
                session_manager.active_session.code,
                session_manager.active_session.number,
                self.is_local_host,
            )
        return self.session_manager

    def is_user_login_type(self) -> bool:
        return True

    def is_drac(self):
        """
        Returns true if the lims used for synchronization of the samples is DRAC
        """
        return self.get_active_lims().name == self.drac.get_lims_name()[0].name

    def set_active_lims(self, lims):
        self.active_lims = lims

    def get_active_lims(self):
        return self.active_lims

    def get_samples(self, lims_id):
        """
        lims_id is the identifier of the lims to be used: ISPyB | DRAC
        """
        logger.debug("[ESRFLIMS] get_samples by lims %s" % lims_id)

        lims_list = [i for i in self.get_lims_name() if i.name == lims_id]
        if len(lims_list) == 1:
            active_lims = lims_list[0]
            logger.debug("[ESRFLIMS] Setting active lims %s" % active_lims.name)
            self.set_active_lims(active_lims)

        logger.debug("[ESRFLIMS] get_samples %s" % self.get_active_lims().name)

        if self.is_drac():
            return self.drac.get_samples(lims_id)
        else:
            return self.ispyb.get_samples(lims_id)

    def get_proposals_by_user(self, login_id: str):
        raise Exception("Not implemented")

    def create_session(self, session_dict):
        pass

    def _store_data_collection_group(self, group_data):
        group_data["sessionId"] = self.ispyb.get_session_id()
        return self.ispyb._store_data_collection_group(
            self._clean_sample_id(group_data)
        )

    def store_data_collection(self, mx_collection, bl_config=None):
        logger.info("Storing datacollection")
        mx_collection["sessionId"] = self.ispyb.get_session_id()

        self.drac.store_data_collection(mx_collection, bl_config)
        return self.ispyb.store_data_collection(
            self._clean_sample_id(mx_collection), bl_config
        )

    def update_data_collection(self, mx_collection):
        logger.info("Updating datacollection")
        mx_collection["sessionId"] = self.ispyb.get_session_id()
        self.drac.update_data_collection(mx_collection)

        return self.ispyb.update_data_collection(self._clean_sample_id(mx_collection))

    def _clean_sample_id(self, mx_collection):
        """
        The sample_id corresponds to the ID in DRAC so when pushing the data
        to ISPyB when DRAC was used we need to remove the id
        """
        mx_collection_copy = mx_collection.copy()
        if self.is_drac():
            if "blSampleId" in mx_collection_copy:
                mx_collection_copy["blSampleId"] = None
                if "sample_reference" in mx_collection_copy:
                    mx_collection_copy["sample_reference"]["blSampleId"] = None
        return mx_collection_copy

    def finalize_data_collection(self, mx_collection):
        logger.info("Storing datacollection")

        mx_collection["sessionId"] = self.ispyb.get_session_id()
        self.drac.finalize_data_collection(mx_collection)
        return self.ispyb.finalize_data_collection(self._clean_sample_id(mx_collection))

    def store_image(self, image_dict):
        self.ispyb.store_image(image_dict)

    def find_sample_by_sample_id(self, sample_id):
        if self.is_drac():
            return self.drac.find_sample_by_sample_id(sample_id)
        return self.ispyb.find_sample_by_sample_id(sample_id)

    def store_robot_action(self, robot_action_dict):
        robot_action_dict["sessionId"] = self.ispyb.get_session_id()
        return self.ispyb.store_robot_action(robot_action_dict)

    def is_session_already_active(self, session_id: str) -> bool:
        return self.drac.is_session_already_active(session_id)

    def set_active_session_by_id(self, session_id: str) -> Session:
        logger.debug("set_active_session_by_id. session_id=%s", str(session_id))

        if self.drac.session_manager.active_session is not None:
            if self.ispyb.session_manager.active_session is not None:
                if self.drac.session_manager.active_session.session_id == session_id:
                    return self.drac.session_manager.active_session

        session = self.drac.set_active_session_by_id(session_id)

        # Check that session is not active already

        if self.ispyb.is_session_already_active_by_code(
            self.drac.session_manager.active_session.code,
            self.drac.session_manager.active_session.number,
        ):
            return self.drac.session_manager.active_session

        if session is not None:
            self.ispyb.get_session_manager_by_code_number(
                self.drac.session_manager.active_session.code,
                self.drac.session_manager.active_session.number,
                self.is_local_host,
            )

            if (
                self.drac.session_manager.active_session is not None
                and self.ispyb.session_manager.active_session is not None
            ):
                logger.info(
                    "[ESRFLIMS] MXCuBE succesfully connected to DRAC:(%s, %s) ISPYB:(%s,%s)"
                    % (
                        self.drac.session_manager.active_session.proposal_name,
                        self.drac.session_manager.active_session.session_id,
                        self.ispyb.session_manager.active_session.proposal_name,
                        self.ispyb.session_manager.active_session.session_id,
                    )
                )
            else:
                logger.exception(
                    "[ESRFLIMS] Problem when set_active_session_by_id. DRAC:(%s) ISPYB:(%s)"
                    % (
                        self.drac.session_manager.active_session.proposal_name,
                        self.ispyb.session_manager.active_session,
                    )
                )
            return self.drac.session_manager.active_session
        else:
            raise Exception("Any candidate session was found")

    def allow_session(self, session: Session):
        return self.drac.allow_session(session)

    def get_session_by_id(self, sid: str):
        return self.drac.get_session_by_id(sid)

    def get_user_name(self):
        return self.drac.get_user_name()

    def get_full_user_name(self):
        return self.drac.get_full_user_name()

    def authenticate(self, login_id: str, password: str) -> LimsSessionManager:
        return self.drac.authenticate(login_id, password)

    def echo(self):
        """Mockup for the echo method."""
        return True

    def is_connected(self):
        return True

    def update_bl_sample(self, bl_sample):
        self.ispyb.update_bl_sample(bl_sample)

    def store_beamline_setup(self, session_id, bl_config):
        self.ispyb.store_beamline_setup(session_id, bl_config)

    def store_energy_scan(self, energyscan_dict):
        energyscan_dict["sessionId"] = self.ispyb.get_session_id()
        self.drac.store_energy_scan(energyscan_dict)
        return self.ispyb.store_energy_scan(self._clean_sample_id(energyscan_dict))

    def store_xfe_spectrum(self, xfespectrum_dict):
        xfespectrum_dict["sessionId"] = self.ispyb.get_session_id()
        self.drac.store_xfe_spectrum(xfespectrum_dict)
        return self.ispyb.store_xfe_spectrum(self._clean_sample_id(xfespectrum_dict))

    def store_workflow(self, *args, **kwargs):
        pass



# -*- coding: utf-8 -*-
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""Resolution concrete implementation.
Example xml file:
<object class="Resolution">
  <username>Resolution</username>
  <actuator_name>resolution</actuator_name>
  <tolerance>1e-4</tolerance>
  <!-- optional -->
  <object href="/pilatus" role="detector"/>
</object>

"""

from mxcubecore.HardwareObjects.abstract.AbstractResolution import AbstractResolution

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class Resolution(AbstractResolution):
    """Resolution as motor"""

    def set_value(self, value, timeout=None):
        """Set the resolution.
        Args:
            value(float): target value [Å]
            timeout(float): optional - timeout [s],
                             if timeout is None: wait forever (default).
        """
        # The precision depends on the difference between the current
        # resolution and the target value - the smaller the difference,
        # the better the precision.
        # We move twice to get the closet possible to the requested resolution.
        super().set_value(value, timeout)
        super().set_value(value, timeout)



# -*- coding: utf-8 -*-
"""Tango Shutter Hardware Object
Example XML:
<object class = "TangoShutter">
  <username>label for users</username>
  <command type="tango" tangoname="my device" name="Open">Open</command>
  <command type="tango" tangoname="my device" name="Close">Close</command>
  <channel type="tango" name="State" tangoname="my device" polling="1000">State</channel>
</object>

"""

import logging
import time

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR


class PX2Guillotine(BaseHardwareObjects.HardwareObject):
    shutterState = {
        # 0:  'ON',
        # 1:  'OFF',
        "False": "CLOSED",
        "True": "OPENED",
        "0": "CLOSED",
        "1": "OPENED",
        "4": "INSERT",
        "5": "EXTRACT",
        "6": "MOVING",
        "7": "STANDBY",
        "8": "FAULT",
        "9": "INIT",
        "10": "RUNNING",
        "11": "ALARM",
        "12": "DISABLED",
        "13": "UNKNOWN",
        "-1": "FAULT",
        "None": "UNKNOWN",
        "UNKNOWN": "UNKNOWN",
        "CLOSE": "CLOSED",
        "OPEN": "OPENED",
        "INSERT": "CLOSED",
        "EXTRACT": "OPENED",
        "MOVING": "MOVING",
        "RUNNING": "MOVING",
        "_": "AUTOMATIC",
        "FAULT": "FAULT",
        "DISABLE": "DISABLED",
        "OFF": "FAULT",
        "STANDBY": "STANDBY",
        "ON": "UNKNOWN",
        "ALARM": "ALARM",
    }
    # shutterState = {
    # None: 'unknown',
    # 'UNKNOWN': 'unknown',
    # 'CLOSE': 'closed',
    # 'OPEN': 'opened',
    # 'INSERT': 'closed',
    # 'EXTRACT': 'opened',
    # 'MOVING': 'moving',
    # 'RUNNING':'moving',
    # '_': 'automatic',
    # 'FAULT': 'fault',
    # 'DISABLE': 'disabled',
    # 'OFF': 'fault',
    # 'STANDBY': 'standby',
    # 'ON': 'unknown'
    # }
    shutterStateString = {
        "ON": "white",
        "OFF": "#012345",
        "CLOSED": "#C03000",
        "CLOSE": "#FF00FF",
        "OPEN": "#00FF00",
        "OPENED": "#00FF00",
        "INSERT": "#EC3CDD",
        "EXTRACT": "#512345",
        "MOVING": "#663300",
        "STANDBY": "#009900",
        "FAULT": "#990000",
        "INIT": "#990000",
        "RUNNING": "#990000",
        "ALARM": "#990000",
        "DISABLED": "#EC3CDD",
        "UNKNOWN": "GRAY",
        "FAULT": "#FF0000",
    }

    def __init__(self, name):
        super().__init__(name)
        logging.info("Guillotine init ")

    def init(self):
        self._shutterStateValue = "UNKNOWN"
        self._currentDistance = "None"
        self._d_security = self.get_property("security_distance")
        self._d_home = self.get_property("safe_distance")
        try:
            self.shutChannel = self.get_channel_object("State")
            self.shutChannel.connect_signal("update", self.shutterStateChanged)

            self.pss = self.get_object_by_role("pss")

            self.connect(
                HWR.beamline.detector.distance,
                "valueChanged",
                self.shutterStateChanged,
            )
            self.connect(
                HWR.beamline.detector.distance,
                "valueChanged",
                self.updateDetectorDistance,
            )

            for command_name in ("_Insert", "_Extract"):
                setattr(self, command_name, self.get_command_object(command_name))

        except KeyError:
            logging.getLogger().warning("%s: cannot report State", self.name())

        try:
            self.pss_door = self.get_property("tangoname_pss")
        except Exception:
            logging.getLogger("HWR").error(
                "Guillotine I11-MA-CE/PSS/DB_DATA: tangopssDevice is not defined "
            )

        if self.pss_door is not None:
            self.memIntChan = self.get_channel_object("memInt")
            self.connect(self.memIntChan, "update", self.updateGuillotine)
        else:
            logging.getLogger("HWR").error("Guillotine: tangopssDevice is not defined ")

    def shutterStateChanged(self, value):
        #
        # emit signal
        #
        if isinstance(value, float):
            value = self.shutChannel.value
        self.shutterStateValue = str(value)
        self.emit("shutterStateChanged", (self.getShutterState(),))

    def getShutterState(self):
        # A tester dans les conditio
        # if self._currentDistance < self._d_security:
        # if self.pss.getWagoState() != "ready":
        #    return "disabled"
        if not self.checkDistance():
            return "disabled"
        return PX2Guillotine.shutterState.get(self.shutterStateValue, "UNKNOWN").lower()

    def updateDetectorDistance(self, value):
        logging.info("UpdateDetectorDistance")
        self._currentDistance = value  # self.detector_distance.res2dist(value)

    def moveGuillotine(self, state):
        if state == "Transfer":
            self.goToSecurityDistance()
        if state == "Collect":
            HWR.beamline.detector.distance.set_value(180)

    def updateGuillotine(self, value):
        # if open door close guillotine but test distance
        # if distance security ok else move detector to security distance
        # wait until distance is reached
        # if self.pss.getWagoState() != "ready":

        if not value:
            if self.checkDistance():
                self._Insert()
            else:
                self.goToSecurityDistance()
                # self.detector_distance.set_value(self._d_home)
                # time.sleep(1.0)# wait distance minimum to insert guillotine
                # self._Insert()

    def modeCollect(self):
        currentDistance = self._currentDistance
        logging.info(
            "PX2Guillotine - setting gonio currentPosition is   %s" % currentDistance
        )
        if self.isInsert():
            if not self.checkDistance():
                HWR.beamline.detector.distance.set_value(self._d_security)
                time.sleep(2.0)
                while HWR.beamline.detector.distance.motorIsMoving():
                    time.sleep(0.5)
                self._Extract()
                time.sleep(0.2)
                HWR.beamline.detector.distance.set_value(currentDistance)
                time.sleep(2.0)
                while HWR.beamline.detector.distance.motorIsMoving():
                    time.sleep(0.5)
            else:
                self._Extract()
                time.sleep(0.2)

    def checkDistance(self):
        logging.info("Current distance is %s" % self._currentDistance)
        if self._currentDistance < self._d_security:
            return False
        else:
            return True

    def isInsert(self):
        if str(self.shutChannel.value) == "INSERT":
            return True
        else:
            return False

    def goToSecurityDistance(self):
        if self._currentDistance < self._d_home:
            HWR.beamline.detector.distance.set_value(self._d_home)
        if str(self.shutChannel.value) == "EXTRACT":
            self._Insert()
        while HWR.beamline.detector.distance.motorIsMoving():
            time.sleep(0.5)

    def isShutterOk(self):
        return not self.getShutterState() in (
            "OFF",
            "UNKNOWN",
            "MOVING",
            "FAULT",
            "INSERT",
            "EXTRACT",
            "INIT",
            "DISABLED",
            "ERROR",
            "ALARM",
            "STANDBY",
        )

    def openShutter(self):
        logging.info("Guillotine extract")
        # if self.checkDistance():
        self._Extract()
        #    return
        logging.getLogger("user_level_log").error(
            " Detector is too closed to the gonio Guillotine couldn't be moved"
        )

    def closeShutter(self):
        logging.info("Guillotine insert")
        # if self.checkDistance():
        self._Insert()
        #    return
        logging.getLogger("user_level_log").error(
            " Detector is too closed to the gonio Guillotine couldn't be moved"
        )

    def setIn(self):
        self._Insert()

    def setOut(self):
        self._Extract()


def test_hwo(hwo):
    hwo.openShutter()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLTableMotor"""

import atexit
import logging
import socket
import time

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "Motor"


class EMBLTableMotor(AbstractMotor):
    """
    EMBLTableMotor defines socket interface to positioning table
    """

    def __init__(self, name):
        AbstractMotor.__init__(self, name)

        self.direction = None
        self.socket = None
        self.enabled = False

        atexit.register(self.close)

    def init(self):
        """
        Creates socket interface
        :return:
        """
        self.direction = self.get_property("direction")
        self.set_position(0)
        self.update_state(self.motor_states.READY)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect(("10.14.90.12", 701))
        self.socket.send("enable (0,1,2,3)\r")
        self.enabled = True

    def is_ready(self):
        """
        Returns ready
        :return: always True
        """
        return True

    def connected(self):
        """
        Sets to ready
        :return: None
        """
        self.set_ready(True)

    def disconnected(self):
        """
        Sets not to ready
        :return:
        """
        self.set_ready(False)

    def stop(self):
        """
        Stops motor movement
        :return: None
        """
        self.socket.send("disable (0,1,2,3)\r")
        self.enabled = False

    def set_value_relative(self, relative_position, wait=False, timeout=None):
        """
        Moves motor by a relative step
        :param relative_position: float
        :param wait: boolean
        :param timeout: in seconds (int)
        :return: None
        """
        self.update_state(self.motor_states.MOVING)
        if not self.enabled:
            self.socket.send("enable (0,1,2,3)\r")
            time.sleep(1)
        if self.direction == "vertical":
            self.socket.send("PTP/r (1), %f\r" % relative_position)
        else:
            self.socket.send("PTP/r (3), %f\r" % relative_position)
        self.update_state(self.motor_states.READY)

    def close(self):
        """
        Closes the socket connection
        :return: None
        """
        try:
            self.socket.close()
            logging.getLogger("HWR").info("EMBLTableMotor: Socket closed")
        except Exception:
            logging.getLogger("HWR").error("EMBLTableMotor: Failed to close the socket")



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from datetime import datetime

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import SimpleHTML

# try:
#    import pdfkit
# except Exception:
#    logging.getLogger("HWR").warning("pdfkit not available")


__credits__ = ["EMBL Hamburg"]
__version__ = "2.3."
__category__ = "General"


TEST_DICT = {
    "ppu": "PPU control",
    "focusing": "Focusing modes",
    "centerbeam": "Beam centering",
}

TEST_COLORS_TABLE = {False: "#FFCCCC", True: "#CCFFCC"}
TEST_COLORS_FONT = {False: "#FE0000", True: "#007800"}


class EMBLBeamlineTest(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.ready_event = None
        self.user_clicked_event = None
        self.test_queue_dict = None
        self.current_test_procedure = None
        self.beamline_name = None
        self.test_directory = None
        self.test_filename = None

        self.available_tests_dict = {}
        self.startup_test_list = []
        self.results_list = None
        self.results_html_list = None

        self.bl_hwobj = None
        self.beam_focusing_hwobj = None

    def init(self):
        """Reads config xml, initiates all necessary hwobj, channels and cmds"""
        self.ready_event = gevent.event.Event()

        self.bl_hwobj = self.get_object_by_role("beamline_setup")
        self.test_filename = "mxcube_test_report"

        try:
            for test in eval(self.get_property("available_tests", "[]")):
                self.available_tests_dict[test] = TEST_DICT[test]
        except Exception:
            logging.getLogger("HWR").debug(
                "BeamlineTest: No test define in xml. "
                + "Setting all tests as available."
            )
        if self.available_tests_dict is None:
            self.available_tests_dict = TEST_DICT

        if self.get_property("startup_tests"):
            self.startup_test_list = eval(self.get_property("startup_tests"))

        if self.get_property("run_tests_at_startup") == True:
            gevent.spawn_later(5, self.start_test_queue, self.startup_test_list)

    def start_test_queue(self, test_list, create_report=True):
        """Runs a list of tests

        :param test_list: list of tests
        :type test_list: list of str
        :param create_report: create html and pdf reports
        :type create_report: bool
        """
        if create_report:
            try:
                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating directory %s" % self.test_directory
                )
                if not os.path.exists(self.test_directory):
                    os.makedirs(self.test_directory)

                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating source directory %s" % self.test_directory
                )
                if not os.path.exists(self.test_directory):
                    os.makedirs(self.test_directory)
            except Exception:
                logging.getLogger("HWR").warning(
                    "BeamlineTest: Unable to create test directories"
                )
                return

        self.results_list = []
        self.results_html_list = []
        for test_index, test_name in enumerate(test_list):
            test_method_name = "test_" + test_name.lower()
            if hasattr(self, test_method_name):
                if test_name in TEST_DICT:
                    logging.getLogger("HWR").debug(
                        "BeamlineTest: Executing test %s (%s)"
                        % (test_name, TEST_DICT[test_name])
                    )

                    progress_info = {
                        "progress_total": len(test_list),
                        "progress_msg": "executing %s" % TEST_DICT[test_name],
                    }
                    self.emit("testProgress", (test_index, progress_info))

                    start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    # self.current_test_procedure = gevent.spawn(\
                    test_result = getattr(self, test_method_name)()
                    # test_result = self.current_test_procedure.get()

                    self.ready_event.wait()
                    self.ready_event.clear()
                    end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    self.results_list.append(
                        {
                            "short_name": test_name,
                            "full_name": TEST_DICT[test_name],
                            "result_bit": test_result.get("result_bit", False),
                            "result_short": test_result.get("result_short", ""),
                            "start_time": start_time,
                            "end_time": end_time,
                        }
                    )

                    self.results_html_list.append(
                        "<h2 id=%s>%s</h2>" % (test_name, TEST_DICT[test_name])
                    )
                    self.results_html_list.append("Started: %s<br>" % start_time)
                    self.results_html_list.append("Ended: %s<br>" % end_time)
                    if test_result.get("result_short"):
                        self.results_html_list.append(
                            "<h3><font color=%s>Result : %s</font></h3>"
                            % (
                                TEST_COLORS_FONT[test_result["result_bit"]],
                                test_result["result_short"],
                            )
                        )
                    if len(test_result.get("result_details", [])) > 0:
                        self.results_html_list.append("<h3>Detailed results:</h3>")
                        self.results_html_list.extend(
                            test_result.get("result_details", [])
                        )
                    self.emit("progressStop", ())
            else:
                msg = (
                    "<h2><font color=%s>Execution method %s "
                    + "for the test %s does not exist</font></h3>"
                )
                self.results_html_list.append(
                    msg
                    % (TEST_COLORS_FONT[False], test_method_name, TEST_DICT[test_name])
                )
                logging.getLogger("HWR").error(
                    "BeamlineTest: Test method " + "%s not available" % test_method_name
                )
            self.results_html_list.append("</p>\n<hr>")

        html_filename = None
        if create_report:
            html_filename = (
                os.path.join(self.test_directory, self.test_filename) + ".html"
            )
            self.generate_report()

        self.emit("testFinished", html_filename)

    def focusing_mode_changed(self, focusing_mode, beam_size):
        """Reemits focusing changed signal

        :param focusing_mode: focusing mode
        :type focusing_mode: str
        :param beam_size: beam size in microns
        :type beam_size: list with two int
        """
        self.emit("focusingModeChanged", focusing_mode, beam_size)

    def get_focus_mode_names(self):
        """Returns available focusing mode names"""
        if self.beam_focusing_hwobj:
            return self.beam_focusing_hwobj.get_focus_mode_names()

    def get_focus_motors(self):
        """Returns focusing motor hwobj"""
        if self.beam_focusing_hwobj is not None:
            return self.beam_focusing_hwobj.get_focus_motors()

    def get_focus_mode(self):
        """Returns active focusing mode"""
        if self.beam_focusing_hwobj is not None:
            return self.beam_focusing_hwobj.get_active_focus_mode()
        else:
            return "Collimated", None

    def set_focus_mode(self, mode):
        """Sets focusing mode

        :param mode: selected focusing mode
        :type mode: str
        """
        if self.beam_focusing_hwobj is not None:
            self.beam_focusing_hwobj.set_focus_mode(mode)

    def set_motor_focus_mode(self, motor, mode):
        """Sets focusing mode to a single motor

        :param motor: motor name
        :type motor: str
        :param mode: focusing mode
        :type mode: str
        """
        if self.beam_focusing_hwobj is not None:
            self.beam_focusing_hwobj.set_motor_focus_mode(motor, mode)

    def ppu_status_changed(self, is_error, text):
        """Reemits ppu status changed signal

        :param is_error: is error
        :type is_error: bool
        :param text: error message
        :type text: str
        """
        self.emit("ppuStatusChanged", (is_error, text))

    def ppu_restart_all(self):
        """Restart ppu processes"""
        if self.bl_hwobj.ppu_control_hwobj is not None:
            self.bl_hwobj.ppu_control_hwobj.restart_all()

    def test_ppu(self):
        """Test ppu"""
        result = {}
        if self.bl_hwobj.ppu_control_hwobj is not None:
            is_error, msg = self.bl_hwobj.ppu_control_hwobj.get_status()
            result["result_bit"] = not is_error
            if result["result_bit"]:
                result["result_short"] = "Test passed"
            else:
                result["result_short"] = "Test failed"

            msg = msg.replace("\n", "\n<br>")
            result["result_details"] = msg.split("\n")
        else:
            result["result_bit"] = False
            result["result_short"] = "Test failed (ppu hwobj not define)."

        self.ready_event.set()

        return result

    def test_focusing(self):
        """Tests focusing"""
        result = {}
        result["result_details"] = []

        active_mode, beam_size = self.get_focus_mode()
        if active_mode is None:
            result["result_bit"] = False
            result["result_short"] = "No focusing mode detected"
        else:
            result["result_bit"] = True
            result["result_short"] = "%s mode detected" % active_mode

        focus_modes = self.get_focus_mode_names()
        focus_motors_list = self.get_focus_motors()

        table_cells = []
        if focus_motors_list:
            for motor in focus_motors_list:
                table_row = []
                table_row.append(motor["motorName"])
                for focus_mode in focus_modes:
                    res = focus_mode in motor["focMode"]
                    table_row.append(
                        "<td bgcolor=%s>%.3f/%.3f</td>"
                        % (
                            TEST_COLORS_TABLE[res],
                            motor["focusingModes"][focus_mode],
                            motor["position"],
                        )
                    )
                table_cells.append(table_row)

        focus_modes = ["Motors"] + list(focus_modes)
        result["result_details"] = SimpleHTML.create_table(focus_modes, table_cells)

        self.ready_event.set()

        return result

    def measure_flux(self):
        """Measures intensity"""
        self.bl_hwobj.flux_hwobj.measure_flux()

    def stop_comm_process(self):
        """Stops pinging"""
        if self.current_test_procedure:
            self.current_test_procedure.kill()
            self.ready_event.set()

    def get_available_tests(self):
        """Returns a list with available tests"""
        return self.available_tests_dict

    def get_startup_test_list(self):
        """Returns a list with tests defined at startup"""
        test_list = []
        for test in self.startup_test_list:
            if TEST_DICT.get(test):
                test_list.append(TEST_DICT[test])
        return test_list

    def generate_report(self):
        """Generates html and pdf report"""
        html_filename = os.path.join(self.test_directory, self.test_filename) + ".html"
        pdf_filename = os.path.join(self.test_directory, self.test_filename) + ".pdf"
        archive_filename = os.path.join(
            self.test_directory,
            datetime.now().strftime("%Y_%m_%d_%H") + "_" + self.test_filename,
        )

        try:
            output_file = open(html_filename, "w")
            output_file.write(SimpleHTML.create_html_start("Beamline test summary"))
            output_file.write("<h1>Beamline %s Test results</h1>" % self.beamline_name)
            output_file.write("<h2>Executed tests:</h2>")
            table_cells = []
            for test in self.results_list:
                table_cells.append(
                    [
                        "bgcolor=%s" % TEST_COLORS_TABLE[test["result_bit"]],
                        "<a href=#%s>%s</a>" % (test["short_name"], test["full_name"]),
                        test["result_short"],
                        test["start_time"],
                        test["end_time"],
                    ]
                )

            table_rec = SimpleHTML.create_table(
                ["Name", "Result", "Start time", "End time"], table_cells
            )
            for row in table_rec:
                output_file.write(row)
            output_file.write("\n<hr>\n")

            for test_result in self.results_html_list:
                output_file.write(test_result + "\n")

            output_file.write(SimpleHTML.create_html_end())
            output_file.close()

            self.emit("htmlGenerated", html_filename)
            logging.getLogger("HWR").info(
                "BeamlineTest: Test result written in file %s" % html_filename
            )
        except Exception:
            logging.getLogger("HWR").error(
                "BeamlineTest: Unable to generate html report file %s" % html_filename
            )

        # try:
        #    pdfkit.from_url(html_filename, pdf_filename)
        #    logging.getLogger("GUI").info("PDF report %s generated" % pdf_filename)
        # except Exception:
        #    logging.getLogger("HWR").error(
        #        "BeamlineTest: Unable to generate pdf report file %s" % pdf_filename
        #    )

        self.emit("testFinished", html_filename)

    def get_result_html(self):
        """Returns html filename"""
        html_filename = os.path.join(self.test_directory, self.test_filename) + ".html"
        if os.path.exists(html_filename):
            return html_filename

    def test_method(self):
        print("Test")



#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.


from mxcubecore.HardwareObjects.abstract.AbstractSlits import AbstractSlits

__credits__ = ["MXCuBE collaboration"]


class LNLSSlits(AbstractSlits):
    def __init__(self, *args):
        AbstractSlits.__init__(self, *args)

    def init(self):
        # Slits start wide open
        self._value = [1.00, 1.00]
        self._min_limits = [0.001, 0.001]
        self._max_limits = [1, 1]

    def set_horizontal_gap(self, value):
        self._value[0] = value
        self.emit("valueChanged", self._value)

    def set_vertical_gap(self, value):
        self._value[1] = value
        self.emit("valueChanged", self._value)

    def stop_horizontal_gap_move(self):
        return

    def stop_vertical_gap_move(self):
        return
